@@ -1,3 +1,5 @@
+import dataclasses
+
 import hashlib
 import html
 import inspect
@@ -9,9 +11,63 @@ import typing
 from markupsafe import Markup
 from pprint import pformat
 from starlette.datastructures import URL
+from starlette.middleware.errors import ServerErrorMiddleware
 from starlette.requests import Request
 from starlette.responses import HTMLResponse, PlainTextResponse, Response
 
+_editor: str = 'none'
+_theme: typing.Literal['light', 'dark'] = 'light'
+open_link_templates: typing.Dict[str, str] = {
+    'none': 'file://{path}',
+    'vscode': 'vscode://file/{path}:{lineno}',
+}
+
+
+def set_editor(name: str) -> None:
+    """
+    Set code editor.
+
+    We will use it to generate file open links. Built-in editors: vscode.
+    """
+    global _editor
+    _editor = name
+
+
+def set_theme(theme: typing.Literal['light', 'dark']) -> None:
+    global _theme
+    _theme = theme
+
+
+def add_link_template(editor: str, template: str) -> None:
+    """
+    Add open file link template. The template accepts two format keys: path and
+    lineno.
+
+    Example:
+        add_link_template('vscode', 'vscode://file/{path}:{lineno}')
+    """
+    open_link_templates[editor] = template
+
+
+def to_ide_link(path: str, lineno: int) -> str:
+    """Generate open file link for current editor."""
+    template = open_link_templates.get(_editor, 'none')
+    return template.format(path=path, lineno=lineno)
+
+
+def install_error_handler() -> None:
+    """
+    Replace Starlette debug exception handler in-place.
+
+    May be, someday, we won't need it.
+    See https://github.com/encode/starlette/discussions/1867
+    """
+
+    def bound_handler(self: ServerErrorMiddleware, request: Request, exc: Exception) -> Response:
+        return exception_handler(request, exc)
+
+    setattr(ServerErrorMiddleware, 'debug_response', bound_handler)
+
 
 def get_relative_filename(path: str) -> str:
     for sys_path in reversed(sorted(sys.path)):
@@ -107,12 +163,36 @@ def mask_secrets(value: str, key: str) -> str:
     return value
 
 
+def highlight(value: str, filename: str) -> str:
+    try:
+        from pygments import highlight
+        from pygments.formatters import HtmlFormatter
+        from pygments.lexers import CssLexer, HtmlLexer, JavascriptLexer, PythonLexer
+
+        style = 'xcode' if _theme == 'light' else 'nord'
+        *_, extension = os.path.splitext(filename)
+        mapping = {
+            '.py': PythonLexer(),
+            '.htm': HtmlLexer(),
+            '.html': HtmlLexer(),
+            '.css': CssLexer(),
+            '.js': JavascriptLexer(),
+        }
+        if lexer := mapping.get(extension):
+            return highlight(value, lexer, HtmlFormatter(noclasses=True, nowrap=True, style=style))  # type: ignore
+        return value
+    except ImportError:
+        return value
+
+
 jinja = jinja2.Environment(loader=jinja2.PackageLoader(__name__))
 jinja.filters.update(
     {
         'symbol': get_symbol,
         'frame_id': frame_id,
         'is_vendor': is_vendor,
+        'highlight': highlight,
+        'to_ide_link': to_ide_link,
         'mask_secrets': mask_secrets,
         'package_dir': get_package_dir,
         'package_name': get_package_name,
@@ -137,17 +217,42 @@ def generate_plain_text(exc: Exception) -> str:
     return "".join(traceback.format_exception(type(exc), exc, exc.__traceback__))
 
 
-def generate_html(request: Request, exc: Exception, limit: int = 7) -> str:
+@dataclasses.dataclass
+class StackItem:
+    exc: Exception
+    frames: typing.List[inspect.FrameInfo]
+    solution: str
+
+
+def generate_html(request: Request, exc: Exception, limit: int = 15) -> str:
     traceback_obj = traceback.TracebackException.from_exception(exc, capture_locals=True)
+    stack = [
+        StackItem(
+            exc=exc,
+            solution=getattr(exc, 'solution', ''),
+            frames=inspect.getinnerframes(exc.__traceback__, limit) if exc.__traceback__ else [],
+        )
+    ]
+    exception = exc
+    while cause := getattr(exception, '__cause__'):
+        stack.append(
+            StackItem(
+                exc=cause,
+                solution=getattr(cause, 'solution', ''),
+                frames=inspect.getinnerframes(cause.__traceback__, limit) if cause.__traceback__ else [],
+            )
+        )
+        exception = cause
 
     template = jinja.get_template('index.html')
     return template.render(
         {
+            'theme': _theme,
             'exception_class': format_qual_name(traceback_obj.exc_type),
             'error_message': str(exc) or '""',
+            'stack': stack,
             'request_method': request.method,
             'request_path': request.url.path,
-            'frames': inspect.getinnerframes(exc.__traceback__, limit) if exc.__traceback__ else [],
             'request_info': {
                 "Method": request.method,
                 "Path": request.url.path,
